#
# Portable Hole Example Product
#
# (C) Copyright 2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Did you read that above bit? This messes with traversal, so don't call me I'll call you.
#
# Author: Andrew Milton <akm@theinternet.com.au>
#
# Ok maybe you can email me to tell me you like it.... but only then...
#

import string # we want string.join and string.split
 
from Globals import DTMLFile
from OFS.Folder import Folder # We're a Folder so we subclass Folder
from Persistence import Persistent # We play with some Persistence..
from OFS.ObjectManager import ObjectManager # We use this for getting things..
from Acquisition import aq_base
from types import TupleType
from ComputedAttribute import ComputedAttribute

def manage_addPortableHole(self, id, title, sourceType, source='', objects=[], REQUEST={}):
	""" Add a new PortableHole """

	#
	# Since our constructor forms are variable, we need to explicitly check for
	# sanity here.
	#
	if sourceType=='folder' and not source:
		raise AttributeError, "source"
	if sourceType=='object' and not objects:
		raise AttributeError, "objects"

	#
	# Make a new one
	#
	o = PortableHole(id, title, sourceType, source, objects)

	# Put it in the ZODB
	self._setObject(id, o)

	# Get it back with an aquisition wrapper
	o = getattr(self, id)
	
	# our object now has an acquisition wrapper, we can do things we can't do in the
	# constructor.
	o.postInitialise()

	# Check for request, if we have one return a management screen
	# otherwise return empty, to allow us to be added from within another product.
	if REQUEST:
		return self.manage_main(self, REQUEST, update_menu=1)
	return ''
	
manage_addPortableHoleForm=DTMLFile('manage_addPortableHoleForm', globals())
manage_addPortableHoleForm2=DTMLFile('manage_addPortableHoleForm2', globals())
manage_showTree_one=DTMLFile('manage_showTree_one', globals())
manage_showTree_many=DTMLFile('manage_showTree_many', globals())
class PortableHole(Folder):
	""" A PortableHole """

	meta_type='PortableHole'
	icon='/misc_/PortableHole/PortableHole_icon.gif'
	#
	# We set this to an empty tuple, because we don't want to add anything
	# inside the PortableHole itself.
	#
	all_meta_types=()

	#
	# _v_ attributes are volatile, they aren't stored in the ZODB, and they die
	# when we're flushed from the in memory cache. This is why there's lots of
	# checks to make sure it's valid.
	#
	# Only used for "Mirror Whole Folder" mode.
	#
	_v_Source=None

	manage_showTree_one=manage_showTree_one
	manage_showTree_many=manage_showTree_many
	manage_editPortableHoleForm=DTMLFile('manage_editPortableHoleForm',
										 globals())

	#
	# We want all the normal folder options, plus one to repoint the
	# PortableHole
	#
	manage_options=(
		Folder.manage_options+
		(
		{'label': 'Repoint', 'action': 'manage_editPortableHoleForm'},
		)
		)
	#
	# We need to setup additional permissions for our new tab, since anything
	# with manage_ in front of it is protected by default
	#
	# Excuse all the commas, AccessControl gets shirty if you don't have the
	# right number of elements in your tuples.
	#
	# These things are case sensitive, one wrong capital can ruin your whole
	# day.
	#
	__ac_permissions__=(
		('View management screens', ('manage_editPortableHoleForm',
									 'manage_editPortableHole'), ('Manager'),)+
		Folder.__ac_permissions__,
		)

	def __init__(self, id, title, sourceType, source='', objects=[]):
		self.id=id
		self.title=title
		self.sourceType=sourceType
		self.source=source
		self.objects=objects
		self.robjects=[]
		self.rpaths={}
		self.initialised=0

	#
	# We try to do things in the right order here, because we should, even
	# though raising an exception should reverse any changes we made due to
	# problems and put the PortableHole back into a known good state.
	#
	# Repointing isn't exactly a good idea in practise.
	#
	def manage_editPortableHole(self, title='', source='', objects=[],
								REQUEST={}):
		""" Edit the PortableHole """
		self.title=title
		if self.sourceType=='folder':
			if not source:
				raise AttributeError, "source"
			if string.find(self.weAre, source)==0:
				raise AttributeError, "Link Loop: %s"%(source)
			self.source=source
			self._v_Source=self.getRealObject(self.source)
		if self.sourceType=='object':
			if not objects:
				raise AttributeError, "objects"
			self.robjects=[]
			self.rpaths=[]
			self.setupObjectsFromPaths(objects)
			self.objects=objects
		if REQUEST:
			return self.manage_main(self, REQUEST, update_menu=1)
		return ''
			
	#
	# Utility function that checks all out objects to make sure we don't get
	# into a loop (ever seen a mirror in a mirror?)
	# Then it adds the good ones a list which we use to pretend they are
	# underneath us.
	#
	def setupObjectsFromPaths(self, objects):
		for op in objects:
			if string.find(self.weAre, op)==0:
				raise AttributeError, "Link Loop: %s"%(op)
			o = self.getRealObject(op)
			if o.getId() in self.robjects:
				raise AttributeError, "ID Clash: %s"%(o.id)
			self.robjects.append(o.getId())
			self.rpaths[o.getId()]=op
		self._p_changed=1

	#
	# Auxillary setup stuff we can't always do without acqusition
	#

	def postInitialise(self):
		#
		# We get our physical path in the zodb, so we can check for loops
		#
		self.weAre=string.join(self.getPhysicalPath(), '/')

		#
		# When we're created some scary stuff goes on with methods being called
		# before we're necessarily ready. I'm not sure why our objects methods
		# would be called at all, some kind of ZODB voodoo I'm sure to make
		# sure we didn't add something before we were wrapped.
		#
		self.initialised=1
		
		if self.sourceType=='folder':
			if string.find(self.weAre, self.source)==0:
				raise AttributeError, "Link Loop: %s"%(self.source)
			self._v_Source=self.getRealObject(self.source)
		if self.sourceType=='object':
			self.setupObjectsFromPaths(self.objects)

	#
	# If we give it an absolute path, it uses the internal Zope mechanism to
	# get us our object.
	#
	def getRealObject(self, sourcePath):
		if not self.initialised:
			return None
		app = self.getPhysicalRoot()
		ob = app.unrestrictedTraverse(string.split(sourcePath,'/'))
		return ob.__of__(self)
		
	# Below here we start overriding methods from ObjectManager which Folder
	# subclasses We do this to fool Zope into thinking we really have some of
	# these objects. There's really not much going on here, if you want to
	# see how easy it is remove all the support for object mode. You end up
	# with trivial implementations of these that just make calls out to the
	# remote folder.


	# Ok we don't want to delete any objects.
	# uh-uh no we don't.

	def _delOb(self, id):
		pass

	#
	# You can't make things in here either...
	#
	def _setOb(self, id):
		pass

	def manage_beforeDelete(self, item, container):
		pass

	#
	# You can't copy us or move us, we don't want you to be able to put us in
	# a loop by pasting us directly below a source item.
	#
	def _canCopy(self, op=0):
		return 0

	#
	# We don't allow you to paste us anywhere either, just in case you
	# managed to trick the machinery into copying us some how.
	#
	def _notifyCopyOfTo(self, container, op=0):
		raise CopyError, "No way"

	def _ph_objects(self):
		# We're not ready yet..
		if not self.initialised:
			return ()
		if self.sourceType=='folder':
			if not self._v_Source:
				self._v_Source=self.getRealObject(self.source)
			if self._v_Source:
				# Just get the real Folder to tell us it's list of objects..
				# Too easy
				return self._v_Source._objects
			return ()
		else:
			# hrm we have to build our own list. We don't want to keep a
			# Persistent list of those objects around (they're in the ZODB,
			# and they could have changed since last time, and that gets
			# messy, so we just get them again.
			x=[]
			for k in self.rpaths.keys():
				x.append(self.getRealObject(self.rpaths[k]))
			return x

	#
	# Steve Alexander suggested this simplification to overriding everything.
	#
	_objects = ComputedAttribute(_ph_objects, 1)

	#
	# Get the object instance.
	#
	def _getOb(self, id, default=[]):
		if not self.initialised:
			return None
		if self.sourceType=='folder':
			if not self._v_Source:
				self._v_Source=self.getRealObject(self.source)
			if self._v_Source:
				# Tell the remote Folder to give it to us, still too easy.
				return self._v_Source._getOb(id, default)
			return None
		elif self.sourceType=='object':
			# Whoops, we have to make one ourselves.. no problem.
			return self.getRealObject(self.rpaths[id])

	def __getattr__(self, name):
		if self.sourceType != 'folder':
			return self.__oldgetattr__(name)
		try:
			if self.sourceType=='folder':
				if not self._v_Source:
					app = Folder.getPhysicalRoot(self)
					self._v_Source=app.unrestrictedTraverse(
						string.split(sourcePath,'/'))
				return self._v_Source.__dict__[name]
		except:
			pass

		try:
			return self.__oldgetattr__(name)
		except:
			pass

		try:
			return ObjectManager.inheritedAttribute(name).__of__(self)
		except:
			pass

		raise AttributeError, name

	#
	# Portable Hole getattr, to defeat Acqusition
	#
	def ph_getattr(self, name):
		try:
			if self.sourceType=='folder':
				if not self._v_Source:
					self._v_Source=self.getRealObject(self.source)
				if self._v_Source:
					return getattr(self._v_Source, name)
			elif self.sourceType=='object' and name[0] != '_':
				if self.rpaths.has_key(name):
					return self.getRealObject(self.rpaths[name])
		except:
			pass
		if self.sourceType=='folder':
			try:
				return self.__oldgetattr__(name)
			except:
				pass
			return self.inheritedAttribute(ObjectManager, name).__of__(self)
		else:
			try:
				return getattr(self, name)
			except:
				pass
			
	#
	# __bobo_traverse__ looks for things, we fix it.
	#
	def __bobo_traverse__(self, REQUEST, name=None):
		try:
			return self.ph_getattr(name)
		except AttributeError:
			pass

		method=REQUEST.get('REQUEST_METHOD', 'GET')
		if not method in ('GET', 'POST'):
			return NullResource(self, name, REQUEST).__of__(self)

		# Waaa. unrestrictedTraverse calls us with a fake REQUEST.
		# There is proabably a better fix for this.
		try:
			REQUEST.RESPONSE.notFoundError("%s\n%s" % (name, method))
		except AttributeError:
			raise KeyError, name
	try:
		#
		# Acquistion calls getattr directly...
		#	
		try:
			__oldgetattr__ = ObjectManager.__getattr__
		except:
			try:
				__oldgetattr__ = ObjectManager.inheritedAttribute('__getattr__')
			except:
				def __oldgetattr__(self, name):
					raise AttributeError, name
	except:
		import traceback
		traceback.print_exc()
	
